package component.system;
import java.util.*;

/**
 * A port is an entry point or exit point for an object. It is used from within the component 
 * to send out messages, and to receive messages from other components.
 * A port has a unique ID, which, once addigned, never changes. 
 * This id is given by the component system. 
 * A port doesn't know with whom it is connected. 
 * A port knows its own type.
 * A port can map incoming messages to certain objects.
 */
public class Port
  extends AbstractPort
{
   protected MessageHandler handleby;   // incoming messages will be mapped onto
   private   PortAddress  id;            // the identification number of this port
   protected PortAddress  target;        // the target when sending messages 
   private   String       information;   // creation information
   protected MessageQueue outqueue;      // queue of messages to be send, but not connected
   protected MessageQueue inqueue;      
   
   /**
    * Creation
    */
   public Port(String type, Component owner)
     {
	this(type,owner,new InvocationMessageHandler(owner));
     }
   public Port(String type, Component owner, MessageHandler handler)
     {
	super(type,owner);
	ComponentSystem.announcePort(this);
	outqueue=new MessageQueue();
	inqueue=new MessageQueue();
	handleby=handler;
     }
   static public AbstractPortCreator getCreator()
     {
	return new PortCreator();
     }
   
   public void setId(PortAddress i)
     {
	id = i;
     }   
   public PortAddress getId()
     {
	return id;
     }   
   public boolean connected()
     {
	return target!=null;
     }
   
   protected void unlink()
     {
	owner=null;
     }
   
   protected void relink(Component own)
     {
	// set owner to own
	setOwner(own);
	while(!inqueue.isEmpty())
	  receiveMessage(inqueue.pop());
     }
   
   public String getInformation()
     {
	return information;
     }
   public void setInformation(String io)
     {
	information=io;
     }   
   public void handleBy(Component o)
     {
	handleBy(new InvocationMessageHandler(o));
     }
   public void handleBy(MessageHandler o)
     {
	handleby=o;
     }

   /**
    * connecting to a port.
    * If connecting. connect(Port p) is called upon handleby
    * when disconnecting. disconnect(Port p) is called upon handleby
    */
   public void setTarget(PortAddress t)
     {
	if (target == null)
	  {
	     if (t != null)
	       {
		  target = t;
		  owner.connect(this);
		  /**
		   * send all messages which could not be sended 
		   * because of connection not yet up !
		   */
		  while(target!=null && !outqueue.isEmpty())
		    doSendMessage(outqueue.pop());
	       }
	     else
	       Fatal("port: ("+this+") reconnection of null ?");
	  }
	else
	  {
	     if (t == null)
	       {
		  target = null;
		  owner.disconnect(this);
	       }
	     else if (!t.equals(target))
	       {
		  target = null;
		  target = t;
	       }
	     else
	       Fatal("port: reconnection of the same target");
	  }
     }
   
   public void disconnect()
     {
	Message m=new Message();
	m.putField("Invoke","Disconnect");
	// the last message we send is a disconnect
	sendMessage(m);
	setTarget(null);
     }
   public PortAddress getTarget()
     {
	return target;
     }

   /**
    * sending of a message is done in 3 steps
    * 1- somebody asks the port to send a message
    * 2- the port asks the component system if it is okay to send the message
    * 3- the port sends the message
    */
   public void sendMessage(Message m)
     {
	if (owner==null)
	  Fatal("port: sendmessage; owner unknown");
	if (m.getField("Invoke")==null)
	  Fatal("port: sendmessage; invoke unknown");
	owner.filterSendMessage(this,m);
     }
   
   public void doSendMessage(Message m)
     {
	if (target==null)
	  {
	     outqueue.push(m);
	  }
	else
	  {
	     m.putField("Target",target);
	     ComponentSystem.doSendMessage(this,m);
	  }
     }
      
   /**
    * receiving a message is done in 3 stepts
    * 1- somebody asks the port to receive a message
    * 2- the port asks his owner to receive the message
    * 3- the owner calls doReceive
    */
   public void receiveMessage(Message m)
     {
	if (owner==null)
	  inqueue.push(m);
	else
	  owner.filterReceiveMessage(this,m);
     }
   public void doReceiveMessage(Message m)
     {
	if (handleby==null)
	  Fatal("port: handleby == null "+name);
	handleby.handleMessage(this,m);
     }
}

class PortCreator extends AbstractPortCreator
{
   public AbstractPort createPort(String name, Component owner, MessageHandler mh)
     {
	return new Port(name, owner,mh);
     }
}
